EC2_CLOUDFORMATION_TEMPLATE = """\
AWSTemplateFormatVersion: 2010-09-09
Transform: AWS::Serverless-2016-10-31
Description: BentoML load balanced template
Parameters:
    AmazonLinux2LatestAmiId:
        Type: AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>
        Default: {ami_id}
Resources:
    SecurityGroupResource:
        Type: AWS::EC2::SecurityGroup
        Properties:
            GroupDescription: "security group for bentoservice"
            SecurityGroupIngress:
                -
                    IpProtocol: tcp
                    CidrIp: 0.0.0.0/0
                    FromPort: 5000
                    ToPort: 5000
                -
                    IpProtocol: tcp
                    CidrIp: 0.0.0.0/0
                    FromPort: 22
                    ToPort: 22
            VpcId: !Ref Vpc1

    Ec2InstanceECRProfile:
        Type: AWS::IAM::InstanceProfile
        Properties:
            Path: /
            Roles: [!Ref EC2Role]

    EC2Role:
        Type: AWS::IAM::Role
        Properties:
            AssumeRolePolicyDocument:
                Statement:
                    -   Effect: Allow
                        Principal:
                            Service: [ec2.amazonaws.com]
                        Action: ['sts:AssumeRole']
            Path: /
            Policies:
                -   PolicyName: ecs-service
                    PolicyDocument:
                        Statement:
                            -   Effect: Allow
                                Action:
                                    -   'ecr:GetAuthorizationToken'
                                    -   'ecr:BatchGetImage'
                                    -   'ecr:GetDownloadUrlForLayer'
                                Resource: '*'

    LaunchTemplateResource:
        Type: AWS::EC2::LaunchTemplate
        Properties:
            LaunchTemplateName: {template_name}
            LaunchTemplateData:
                IamInstanceProfile:
                    Arn: !GetAtt Ec2InstanceECRProfile.Arn
                ImageId: !Ref AmazonLinux2LatestAmiId
                InstanceType: {instance_type}
                UserData: "{user_data}"
                SecurityGroupIds:
                - !GetAtt SecurityGroupResource.GroupId

    TargetGroup:
        Type: AWS::ElasticLoadBalancingV2::TargetGroup
        Properties:
            VpcId: !Ref Vpc1
            Protocol: HTTP
            Port: 5000
            TargetType: instance
            HealthCheckEnabled: true
            HealthCheckIntervalSeconds: {target_health_check_interval_seconds}
            HealthCheckPath: {target_health_check_path}
            HealthCheckPort: {target_health_check_port}
            HealthCheckProtocol: HTTP
            HealthCheckTimeoutSeconds: {target_health_check_timeout_seconds}
            HealthyThresholdCount: {target_health_check_threshold_count}

    LoadBalancerSecurityGroup:
        Type: AWS::EC2::SecurityGroup
        Properties:
            GroupDescription: "security group for loadbalancing"
            VpcId: !Ref Vpc1
            SecurityGroupIngress:
                -
                    IpProtocol: tcp
                    CidrIp: 0.0.0.0/0
                    FromPort: 80
                    ToPort: 80

    InternetGateway:
        Type: AWS::EC2::InternetGateway

    Gateway:
        Type: AWS::EC2::VPCGatewayAttachment
        Properties:
            InternetGatewayId: !Ref InternetGateway
            VpcId: !Ref Vpc1

    PublicRouteTable:
        Type: AWS::EC2::RouteTable
        Properties:
            VpcId: !Ref Vpc1

    PublicRoute:
        Type: AWS::EC2::Route
        DependsOn: Gateway
        Properties:
            DestinationCidrBlock: 0.0.0.0/0
            GatewayId: !Ref InternetGateway
            RouteTableId: !Ref PublicRouteTable

    RouteTableSubnetTwoAssociationOne:
        Type: AWS::EC2::SubnetRouteTableAssociation
        Properties:
          RouteTableId: !Ref PublicRouteTable
          SubnetId: !Ref Subnet1
    RouteTableSubnetTwoAssociationTwo:
        Type: AWS::EC2::SubnetRouteTableAssociation
        Properties:
          RouteTableId: !Ref PublicRouteTable
          SubnetId: !Ref Subnet2

    Vpc1:
        Type: AWS::EC2::VPC
        Properties:
            CidrBlock: 172.31.0.0/16
            EnableDnsHostnames: true
            EnableDnsSupport: true
            InstanceTenancy: default

    Subnet1:
        Type: AWS::EC2::Subnet
        Properties:
            VpcId: !Ref Vpc1
            AvailabilityZone:
                Fn::Select:
                    - 0
                    - Fn::GetAZs: ""
            CidrBlock: 172.31.16.0/20
            MapPublicIpOnLaunch: true

    Subnet2:
        Type: AWS::EC2::Subnet
        Properties:
            VpcId: !Ref Vpc1
            AvailabilityZone:
                Fn::Select:
                    - 1
                    - Fn::GetAZs: ""
            CidrBlock: 172.31.0.0/20
            MapPublicIpOnLaunch: true

    LoadBalancer:
        Type: AWS::ElasticLoadBalancingV2::LoadBalancer
        Properties:
            IpAddressType: ipv4
            Name: {elb_name}
            Scheme: internet-facing
            SecurityGroups:
                - !Ref LoadBalancerSecurityGroup
            Subnets:
                - !Ref Subnet1
                - !Ref Subnet2
            Type: application

    Listener:
        Type: AWS::ElasticLoadBalancingV2::Listener
        Properties:
            DefaultActions:
                -   Type: forward
                    TargetGroupArn: !Ref TargetGroup
            LoadBalancerArn: !Ref LoadBalancer
            Port: 80
            Protocol: HTTP

    AutoScalingGroup:
        Type: AWS::AutoScaling::AutoScalingGroup
        DependsOn: Gateway
        Properties:
            MinSize: {autoscaling_min_size}
            MaxSize: {autoscaling_max_size}
            DesiredCapacity: {autoscaling_desired_capacity}
            AvailabilityZones:
                - Fn::Select:
                    - 0
                    - Fn::GetAZs: ""
                - Fn::Select:
                    - 1
                    - Fn::GetAZs: ""
            LaunchTemplate:
                LaunchTemplateId: !Ref LaunchTemplateResource
                Version: !GetAtt LaunchTemplateResource.LatestVersionNumber
            TargetGroupARNs:
                - !Ref TargetGroup
            VPCZoneIdentifier:
            - !Ref Subnet1
            - !Ref Subnet2
        UpdatePolicy:
            AutoScalingReplacingUpdate:
                WillReplace: true

Outputs:
    S3Bucket:
        Value: {s3_bucket_name}
        Description: Bucket to store sam artifacts
    AutoScalingGroup:
        Value: !Ref AutoScalingGroup
        Description: Autoscaling group name
    TargetGroup:
        Value: !Ref TargetGroup
        Description: Target group for load balancer
    Url:
        Value: !Join ['', ['http://', !GetAtt [LoadBalancer, DNSName]]]
        Description: URL of the bento service

"""


EC2_USER_INIT_SCRIPT = """MIME-Version: 1.0
Content-Type: multipart/mixed; boundary=\"==MYBOUNDARY==\"

--==MYBOUNDARY==
Content-Type: text/cloud-config; charset=\"us-ascii\"

runcmd:

- sudo yum update -y
- sudo amazon-linux-extras install docker -y
- sudo service docker start
- sudo usermod -a -G docker ec2-user
- curl "https://awscli.amazonaws.com/awscli-exe-linux-x86_64.zip" -o "awscliv2.zip"
- unzip awscliv2.zip
- sudo ./aws/install
- ln -s /usr/bin/aws aws
- aws ecr get-login-password --region {region}|docker login --username AWS --password-stdin {registry}
- docker pull {tag}
- docker run -p {bentoservice_port}:{bentoservice_port} {tag}

--==MYBOUNDARY==--
"""  # noqa: E501



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import re
import shutil
import subprocess
import logging
import sys

import docker

from bentoml.utils.tempdir import TempDirectory
from bentoml.saved_bundle import loader
from bentoml.yatai.deployment.azure_functions.constants import (
    MAX_RESOURCE_GROUP_NAME_LENGTH,
    MAX_STORAGE_ACCOUNT_NAME_LENGTH,
    MAX_FUNCTION_NAME_LENGTH,
    MAX_CONTAINER_REGISTRY_NAME_LENGTH,
    DEFAULT_MIN_INSTANCE_COUNT,
    DEFAULT_MAX_BURST,
    DEFAULT_PREMIUM_PLAN_SKU,
    DEFAULT_FUNCTION_AUTH_LEVEL,
)
from bentoml.yatai.deployment.azure_functions.templates import AZURE_API_FUNCTION_JSON
from bentoml.yatai.deployment.operator import DeploymentOperatorBase
from bentoml.yatai.deployment.docker_utils import ensure_docker_available_or_raise
from bentoml.exceptions import (
    BentoMLException,
    MissingDependencyException,
    AzureServiceError,
    YataiDeploymentException,
)
from bentoml.yatai.proto.deployment_pb2 import (
    ApplyDeploymentResponse,
    DeploymentState,
    DescribeDeploymentResponse,
    DeleteDeploymentResponse,
)
from bentoml.yatai.proto.repository_pb2 import GetBentoRequest
from bentoml.yatai.status import Status
from bentoml.configuration import LAST_PYPI_RELEASE_VERSION

logger = logging.getLogger(__name__)


def _assert_az_cli_logged_in():
    account_info = _call_az_cli(
        command=['az', 'account', 'show'], message='show Azure account'
    )
    if not account_info['user'] or not account_info['homeTenantId']:
        raise YataiDeploymentException(
            'A signed in Azure CLI is required for Azure Function deployment'
        )


def _assert_azure_cli_available():
    try:
        _call_az_cli(command=['az', 'account', 'show'], message='show Azure account')
    except FileNotFoundError:
        raise MissingDependencyException(
            'azure cli is required for this deployment. Please visit '
            'https://docs.microsoft.com/en-us/cli/azure/install-azure-cli '
            'for instructions'
        )


def _init_azure_functions_project(
    azure_functions_project_dir, bento_path, azure_functions_config
):
    try:
        local_bentoml_path = os.path.dirname(__file__)
        shutil.copytree(bento_path, azure_functions_project_dir)
        shutil.copy(
            os.path.join(local_bentoml_path, 'host.json'),
            os.path.join(azure_functions_project_dir, 'host.json'),
        )
        shutil.copy(
            os.path.join(local_bentoml_path, 'local.settings.json'),
            os.path.join(azure_functions_project_dir, 'local.settings.json'),
        )
        shutil.copy(
            os.path.join(local_bentoml_path, 'Dockerfile'),
            os.path.join(azure_functions_project_dir, 'Dockerfile-azure'),
        )

        app_path = os.path.join(azure_functions_project_dir, 'app')
        os.mkdir(app_path)
        shutil.copy(
            os.path.join(local_bentoml_path, 'app_init.py'),
            os.path.join(app_path, '__init__.py'),
        )
        with open(os.path.join(app_path, 'function.json'), 'w') as f:
            f.write(
                AZURE_API_FUNCTION_JSON.format(
                    function_auth_level=azure_functions_config.function_auth_level
                    or DEFAULT_FUNCTION_AUTH_LEVEL
                )
            )
    except Exception as e:
        raise BentoMLException(f'Failed to initialize azure function project. {str(e)}')


def _generate_azure_resource_names(namespace, deployment_name):
    # Generate resource names base on
    # https://docs.microsoft.com/en-us/azure/azure-resource-manager/management/resource-name-rules

    # 1-90, alphanumeric(A-Za-z0-9) underscores, parentheses, hyphens, periods
    # scope subscription
    resource_group_name = f'{namespace}-{deployment_name}'
    if len(resource_group_name) > MAX_RESOURCE_GROUP_NAME_LENGTH:
        resource_group_name = f'{namespace[:29]}-{deployment_name[:60]}'

    # 3-24 a-z0-9, scope: global
    storage_account_name = f'{namespace}{deployment_name}'.lower()
    if len(storage_account_name) > MAX_STORAGE_ACCOUNT_NAME_LENGTH:
        storage_account_name = f'{namespace[:5]}{deployment_name[0:19]}'
    # Replace invalid chars in storage account name to '0'
    storage_account_name = re.sub(re.compile("[^a-z0-9]"), '0', storage_account_name)

    # Azure has no documentation on the requirements for function plan name.
    function_plan_name = deployment_name

    # same as Microsoft.Web/sites
    # 2-60, alphanumeric and hyphens. scope global
    function_name = f'{namespace}-{deployment_name}'
    if len(deployment_name) > MAX_FUNCTION_NAME_LENGTH:
        function_name = f'{namespace[:19]}-{deployment_name[:40]}'
    function_name = re.sub(re.compile("[^a-zA-Z0-9-]"), '-', function_name)

    # 5-50, alphanumeric scope global
    container_registry_name = f'{namespace}{deployment_name}'
    if len(container_registry_name) > MAX_CONTAINER_REGISTRY_NAME_LENGTH:
        container_registry_name = f'{namespace[:10]}{deployment_name[:40]}'
    container_registry_name = re.sub(
        re.compile("[^a-zA-Z0-9]"), '0', container_registry_name
    )

    return (
        resource_group_name,
        storage_account_name,
        function_plan_name,
        function_name,
        container_registry_name,
    )


def _call_az_cli(command, message, parse_json=True):
    proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    sys_default_encoding = sys.getfilesystemencoding()
    stdout, stderr = proc.communicate()
    if proc.returncode == 0:
        result = stdout.decode(sys_default_encoding)
        if result.endswith('\x1b[0m'):
            # remove console color code: \x1b[0m
            # https://github.com/Azure/azure-cli/issues/9903
            result = result.replace('\x1b[0m', '')
        logger.debug(f'AZ command "{" ".join(command)}" Result: {result}')
        if parse_json:
            return json.loads(result)
        else:
            return result
    else:
        error_message = stderr.decode(sys_default_encoding)
        if not error_message:
            error_message = stdout.decode(sys_default_encoding)
        logger.error(f'AZ command: "{" ".join(command)}" failed: {error_message}')
        raise AzureServiceError(f'Failed to {message}. {str(error_message)}')


def _login_acr_registry(acr_name, resource_group_name):
    result = _call_az_cli(
        command=[
            'az',
            'acr',
            'login',
            '--name',
            acr_name,
            '--resource-group',
            resource_group_name,
        ],
        message='log into Azure container registry',
        parse_json=False,
    ).replace('\n', '')
    if result == 'Login Succeed':
        raise AzureServiceError(
            f'Failed to log into Azure container registry. {result}'
        )


def _build_and_push_docker_image_to_azure_container_registry(
    azure_functions_project_dir,
    container_registry_name,
    resource_group_name,
    bento_name,
    bento_version,
    bento_python_version,
):
    _login_acr_registry(container_registry_name, resource_group_name)
    docker_client = docker.from_env()
    major, minor, _ = bento_python_version.split('.')
    try:
        docker_client.ping()
    except docker.errors.APIError as err:
        raise YataiDeploymentException(
            f'Failed to get response from docker server: {str(err)}'
        )
    tag = f'{container_registry_name}.azurecr.io/{bento_name}:{bento_version}'.lower()
    logger.debug(f'Building docker image {tag}')
    try:
        docker_client.images.build(
            path=azure_functions_project_dir,
            dockerfile=os.path.join(azure_functions_project_dir, 'Dockerfile-azure'),
            tag=tag,
            buildargs={
                'BENTOML_VERSION': LAST_PYPI_RELEASE_VERSION,
                'PYTHON_VERSION': major + minor,
            },
        )
        logger.debug('Finished building docker image')
    except docker.errors.BuildError as e:
        raise YataiDeploymentException(
            f'Failed to build docker image. BuildError: {str(e)}'
        )
    except docker.errors.APIError as e:
        raise YataiDeploymentException(
            f'Failed to build docker image. APIError: {str(e)}'
        )
    logger.debug(f'Pushing docker image {tag}')
    try:
        docker_client.images.push(tag)
        logger.debug('Finished pushing docker image')
    except docker.errors.APIError as e:
        raise YataiDeploymentException(
            f'Failed to push docker image. APIError: {str(e)}'
        )

    return tag


def _get_docker_login_info(resource_group_name, container_registry_name):
    _call_az_cli(
        command=[
            'az',
            'acr',
            'update',
            '--name',
            container_registry_name,
            '--admin-enabled',
            'true',
        ],
        message='enable admin for Azure container registry',
    )
    docker_login_info = _call_az_cli(
        command=[
            'az',
            'acr',
            'credential',
            'show',
            '--name',
            container_registry_name,
            '--resource-group',
            resource_group_name,
        ],
        message='show Azure container registry credential info',
    )

    return docker_login_info['username'], docker_login_info['passwords'][0]['value']


def _set_cors_settings(function_name, resource_group_name):
    # To allow all, use `*` and  remove all other origins in the list.
    cors_list_result = _call_az_cli(
        command=[
            'az',
            'functionapp',
            'cors',
            'show',
            '--name',
            function_name,
            '--resource-group',
            resource_group_name,
        ],
        message='show Azure functionapp cors settings',
    )
    for origin_url in cors_list_result['allowedOrigins']:
        _call_az_cli(
            command=[
                'az',
                'functionapp',
                'cors',
                'remove',
                '--name',
                function_name,
                '--resource-group',
                resource_group_name,
                '--allowed-origins',
                origin_url,
            ],
            message=f'remove allowed origin "{origin_url}"from Azure functionapp',
        )

    _call_az_cli(
        command=[
            'az',
            'functionapp',
            'cors',
            'add',
            '--name',
            function_name,
            '--resource-group',
            resource_group_name,
            '--allowed-origins',
            '*',
        ],
        message='update Azure functionapp cors setting',
    )


def _deploy_azure_functions(
    namespace, deployment_name, deployment_spec, bento_pb, bento_path,
):
    azure_functions_config = deployment_spec.azure_functions_operator_config
    (
        resource_group_name,
        storage_account_name,
        function_plan_name,
        function_name,
        container_registry_name,
    ) = _generate_azure_resource_names(namespace, deployment_name)
    with TempDirectory() as temp_dir:
        azure_functions_project_dir = os.path.join(temp_dir, deployment_spec.bento_name)
        _init_azure_functions_project(
            azure_functions_project_dir, bento_path, azure_functions_config,
        )
        _call_az_cli(
            command=[
                'az',
                'group',
                'create',
                '--name',
                resource_group_name,
                '--location',
                azure_functions_config.location,
            ],
            message='create Azure resource group',
        )
        _call_az_cli(
            command=[
                'az',
                'storage',
                'account',
                'create',
                '--name',
                storage_account_name,
                '--resource-group',
                resource_group_name,
            ],
            message='Create Azure storage account',
        )
        _call_az_cli(
            command=[
                'az',
                'functionapp',
                'plan',
                'create',
                '--name',
                function_plan_name,
                '--is-linux',
                '--sku',
                azure_functions_config.premium_plan_sku or DEFAULT_PREMIUM_PLAN_SKU,
                '--min-instances',
                str(azure_functions_config.min_instances)
                or str(DEFAULT_MIN_INSTANCE_COUNT),
                '--max-burst',
                str(azure_functions_config.max_burst) or str(DEFAULT_MAX_BURST),
                '--resource-group',
                resource_group_name,
            ],
            message='create Azure functionapp premium plan',
        )
        # Add note for why choose standard
        _call_az_cli(
            command=[
                'az',
                'acr',
                'create',
                '--name',
                container_registry_name,
                '--sku',
                'standard',
                '--resource-group',
                resource_group_name,
            ],
            message='create Azure container registry',
        )
        try:
            docker_tag = _build_and_push_docker_image_to_azure_container_registry(
                azure_functions_project_dir=azure_functions_project_dir,
                container_registry_name=container_registry_name,
                resource_group_name=resource_group_name,
                bento_name=bento_pb.name,
                bento_version=bento_pb.version,
                bento_python_version=bento_pb.bento_service_metadata.env.python_version,
            )
        except Exception as e:
            raise AzureServiceError(
                f'Failed to create Azure Function docker image. {str(e)}'
            )
        docker_username, docker_password = _get_docker_login_info(
            resource_group_name, container_registry_name
        )
        _call_az_cli(
            command=[
                'az',
                'functionapp',
                'create',
                '--name',
                function_name,
                '--storage-account',
                storage_account_name,
                '--resource-group',
                resource_group_name,
                '--plan',
                function_plan_name,
                '--functions-version',
                '3',
                '--deployment-container-image-name',
                docker_tag,
                '--docker-registry-server-user',
                docker_username,
                '--docker-registry-server-password',
                docker_password,
            ],
            message='create Azure functionapp',
        )
        _set_cors_settings(function_name, resource_group_name)


def _update_azure_functions(
    namespace, deployment_name, deployment_spec, bento_pb, bento_path,
):
    azure_functions_config = deployment_spec.azure_functions_operator_config
    (
        resource_group_name,
        _,
        _,
        function_name,
        container_registry_name,
    ) = _generate_azure_resource_names(namespace, deployment_name)
    with TempDirectory() as temp_dir:
        azure_functions_project_dir = os.path.join(temp_dir, deployment_spec.bento_name)
        _init_azure_functions_project(
            azure_functions_project_dir, bento_path, azure_functions_config,
        )
        docker_tag = _build_and_push_docker_image_to_azure_container_registry(
            azure_functions_project_dir=azure_functions_project_dir,
            container_registry_name=container_registry_name,
            resource_group_name=resource_group_name,
            bento_name=bento_pb.name,
            bento_version=bento_pb.version,
            bento_python_version=bento_pb.bento_service_metadata.env.python_version,
        )
        _call_az_cli(
            command=[
                'az',
                'functionapp',
                'config',
                'container',
                'set',
                '--name',
                function_name,
                '--resource-group',
                resource_group_name,
                '--docker-custom-image-name',
                docker_tag,
            ],
            message='update Azure functionapp settings',
        )


class AzureFunctionsDeploymentOperator(DeploymentOperatorBase):
    def __init__(self, yatai_service):
        super(AzureFunctionsDeploymentOperator, self).__init__(yatai_service)
        ensure_docker_available_or_raise()
        _assert_azure_cli_available()
        _assert_az_cli_logged_in()

    def add(self, deployment_pb):
        try:
            deployment_spec = deployment_pb.spec
            if not deployment_spec.azure_functions_operator_config.location:
                raise YataiDeploymentException(
                    'Azure Functions parameter "location" is missing'
                )
            bento_repo_pb = self.yatai_service.GetBento(
                GetBentoRequest(
                    bento_name=deployment_spec.bento_name,
                    bento_version=deployment_spec.bento_version,
                )
            )
            return self._add(
                deployment_pb, bento_repo_pb.bento, bento_repo_pb.bento.uri.uri
            )
        except BentoMLException as error:
            deployment_pb.state.state = DeploymentState.ERROR
            deployment_pb.state.error_message = f'Error: {str(error)}'
            return ApplyDeploymentResponse(
                status=error.status_proto, deployment=deployment_pb
            )

    def _add(self, deployment_pb, bento_pb, bento_path):
        if loader._is_remote_path(bento_path):
            with loader._resolve_remote_bundle_path(bento_path) as local_path:
                return self._add(deployment_pb, bento_pb, local_path)
        try:
            _deploy_azure_functions(
                deployment_spec=deployment_pb.spec,
                deployment_name=deployment_pb.name,
                namespace=deployment_pb.namespace,
                bento_pb=bento_pb,
                bento_path=bento_path,
            )
            return ApplyDeploymentResponse(status=Status.OK(), deployment=deployment_pb)
        except AzureServiceError as error:
            resource_group_name, _, _, _, _, = _generate_azure_resource_names(
                deployment_pb.namespace, deployment_pb.name
            )
            logger.debug(
                'Failed to create Azure Functions. Cleaning up Azure resources'
            )
            try:
                _call_az_cli(
                    command=[
                        'az',
                        'group',
                        'delete',
                        '-y',
                        '--name',
                        resource_group_name,
                    ],
                    message='delete Azure resource group',
                )
            except AzureServiceError:
                pass
            raise error

    def update(self, deployment_pb, previous_deployment):
        try:
            bento_repo_pb = self.yatai_service.GetBento(
                GetBentoRequest(
                    bento_name=deployment_pb.spec.bento_name,
                    bento_version=deployment_pb.spec.bento_version,
                )
            )
            bento_pb = bento_repo_pb.bento
            return self._update(
                deployment_pb, previous_deployment, bento_pb, bento_pb.uri.uri
            )
        except BentoMLException as error:
            deployment_pb.state.state = DeploymentState.ERROR
            deployment_pb.state.error_message = (
                f'Encounter error when updating Azure Functions deployment: '
                f'{str(error)}'
            )
            return ApplyDeploymentResponse(
                status=error.status_proto, deployment=deployment_pb
            )

    def _update(self, deployment_pb, current_deployment, bento_pb, bento_path):
        if loader._is_remote_path(bento_path):
            with loader._resolve_remote_bundle_path(bento_path) as local_path:
                return self._update(
                    deployment_pb, current_deployment, bento_pb, local_path
                )
        if (
            deployment_pb.spec.bento_name != current_deployment.spec.bento_name
            or deployment_pb.spec.bento_version != current_deployment.spec.bento_version
        ):
            logger.debug(
                'BentoService tag is different from current Azure Functions '
                'deployment, creating new Azure Functions project and push to ACR'
            )
            _update_azure_functions(
                deployment_spec=deployment_pb.spec,
                deployment_name=deployment_pb.name,
                namespace=deployment_pb.namespace,
                bento_pb=bento_pb,
                bento_path=bento_path,
            )
        (
            resource_group_name,
            _,
            function_plan_name,
            _,
            _,
        ) = _generate_azure_resource_names(
            namespace=deployment_pb.namespace, deployment_name=deployment_pb.name
        )
        _call_az_cli(
            command=[
                'az',
                'functionapp',
                'plan',
                'update',
                '--name',
                function_plan_name,
                '--resource-group',
                resource_group_name,
                '--max-burst',
                str(deployment_pb.spec.azure_functions_operator_config.max_burst),
                '--min-instances',
                str(deployment_pb.spec.azure_functions_operator_config.min_instances),
                '--sku',
                deployment_pb.spec.azure_functions_operator_config.premium_plan_sku,
            ],
            message='update Azure functionapp plan',
        )
        return ApplyDeploymentResponse(deployment=deployment_pb, status=Status.OK())

    def delete(self, deployment_pb):
        try:
            resource_group_name, _, _, _, _ = _generate_azure_resource_names(
                namespace=deployment_pb.namespace, deployment_name=deployment_pb.name
            )
            _call_az_cli(
                command=['az', 'group', 'delete', '-y', '--name', resource_group_name],
                message='delete Azure resource group',
                parse_json=False,
            )
            return DeleteDeploymentResponse(status=Status.OK())
        except BentoMLException as error:
            return DeleteDeploymentResponse(status=error.status_proto)

    def describe(self, deployment_pb):
        try:
            (
                resource_group_name,
                _,
                _,
                function_name,
                _,
            ) = _generate_azure_resource_names(
                deployment_pb.namespace, deployment_pb.name
            )
            show_function_result = _call_az_cli(
                command=[
                    'az',
                    'functionapp',
                    'show',
                    '--name',
                    function_name,
                    '--resource-group',
                    resource_group_name,
                ],
                message='show Azure functionapp detail',
            )
            keys = [
                'defaultHostName',
                'enabledHostNames',
                'hostNames',
                'id',
                'kind',
                'lastModifiedTimeUtc',
                'location',
                'name',
                'repositorySiteName',
                'reserved',
                'resourceGroup',
                'state',
                'type',
                'usageState',
            ]
            # Need find more documentation on the status of functionapp. For now, any
            # other status is error.
            if show_function_result['state'] == 'Running':
                state = DeploymentState.RUNNING
            else:
                state = DeploymentState.ERROR
            info_json = {k: v for k, v in show_function_result.items() if k in keys}
            deployment_state = DeploymentState(
                info_json=json.dumps(info_json, default=str), state=state,
            )
            deployment_state.timestamp.GetCurrentTime()
            return DescribeDeploymentResponse(
                state=deployment_state, status=Status.OK()
            )
        except BentoMLException as error:
            return DescribeDeploymentResponse(status=error.status_proto)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import tarfile
import uuid
from datetime import datetime

from simple_di import Provide, inject

from bentoml import __version__ as BENTOML_VERSION
from bentoml.configuration.containers import BentoMLContainer
from bentoml.exceptions import (
    BadInput,
    BentoMLException,
    InvalidArgument,
    YataiRepositoryException,
)
from bentoml.saved_bundle import safe_retrieve
from bentoml.utils import ProtoMessageToDict
from bentoml.utils.docker_utils import (
    to_valid_docker_image_name,
    to_valid_docker_image_version,
)
from bentoml.utils.tempdir import TempDirectory
from bentoml.utils.usage_stats import track
from bentoml.yatai.db import DB
from bentoml.yatai.db.stores.lock import LockStore
from bentoml.yatai.deployment.docker_utils import ensure_docker_available_or_raise
from bentoml.yatai.deployment.operator import get_deployment_operator
from bentoml.yatai.grpc_stream_utils import DownloadBentoStreamResponses
from bentoml.yatai.locking.lock import DEFAULT_TTL_MIN
from bentoml.yatai.locking.lock import LockType, lock
from bentoml.yatai.proto import status_pb2
from bentoml.yatai.proto.deployment_pb2 import (
    ApplyDeploymentResponse,
    DeleteDeploymentResponse,
    DeploymentSpec,
    DescribeDeploymentResponse,
    GetDeploymentResponse,
    ListDeploymentsResponse,
)
from bentoml.yatai.proto.repository_pb2 import (
    AddBentoResponse,
    BentoUri,
    ContainerizeBentoResponse,
    DangerouslyDeleteBentoResponse,
    DownloadBentoResponse,
    GetBentoResponse,
    ListBentoResponse,
    UpdateBentoResponse,
    UploadBentoResponse,
    UploadStatus,
)
from bentoml.yatai.proto.yatai_service_pb2 import (
    GetYataiServiceVersionResponse,
    HealthCheckResponse,
)
from bentoml.yatai.repository.base_repository import BaseRepository
from bentoml.yatai.repository.file_system_repository import FileSystemRepository
from bentoml.yatai.status import Status
from bentoml.yatai.utils import docker_build_logs
from bentoml.yatai.validator import validate_deployment_pb

logger = logging.getLogger(__name__)


def track_deployment_delete(deployment_operator, created_at, force_delete=False):
    operator_name = DeploymentSpec.DeploymentOperator.Name(deployment_operator)
    up_time = int((datetime.utcnow() - created_at.ToDatetime()).total_seconds())
    track(
        f'deployment-{operator_name}-stop',
        {'up_time': up_time, 'force_delete': force_delete},
    )


def is_file_system_repo(repo_instance) -> bool:
    return isinstance(repo_instance, FileSystemRepository)


# NOTES: How do we type hints this function?
def get_yatai_service_impl(base=object):
    # This helps avoid loading YataiServicer grpc file when using local YataiService
    # implementation. This speeds up regular save/load operations when Yatai is not used

    class YataiServiceImpl(base):

        # pylint: disable=unused-argument
        # pylint: disable=broad-except
        @inject
        def __init__(
            self,
            database: DB,
            repository: BaseRepository,
            default_namespace: str = Provide[BentoMLContainer.config.yatai.namespace],
        ):
            self.default_namespace = default_namespace
            self.repo = repository
            self.db = database

        def HealthCheck(self, request, context=None):
            return HealthCheckResponse(status=Status.OK())

        def GetYataiServiceVersion(self, request, context=None):
            return GetYataiServiceVersionResponse(
                status=Status.OK, version=BENTOML_VERSION
            )

        def ApplyDeployment(self, request, context=None):
            deployment_id = f"{request.deployment.name}_{request.deployment.namespace}"
            spec = request.deployment.spec
            bento_id = f"{spec.bento_name}_{spec.bento_version}"
            with lock(
                self.db, [(deployment_id, LockType.WRITE), (bento_id, LockType.READ)]
            ) as (
                sess,
                _,
            ):
                # to bento level lock
                try:
                    # apply default namespace if not set
                    request.deployment.namespace = (
                        request.deployment.namespace or self.default_namespace
                    )

                    validation_errors = validate_deployment_pb(request.deployment)
                    if validation_errors:
                        raise InvalidArgument(
                            'Failed to validate deployment. {errors}'.format(
                                errors=validation_errors
                            )
                        )

                    previous_deployment = self.db.deployment_store.get(
                        sess, request.deployment.name, request.deployment.namespace
                    )
                    if not previous_deployment:
                        request.deployment.created_at.GetCurrentTime()
                    request.deployment.last_updated_at.GetCurrentTime()

                    self.db.deployment_store.insert_or_update(sess, request.deployment)
                    # find deployment operator based on deployment spec
                    operator = get_deployment_operator(self, request.deployment)

                    # deploying to target platform
                    if previous_deployment:
                        response = operator.update(
                            request.deployment, previous_deployment
                        )
                    else:
                        response = operator.add(request.deployment)

                    if response.status.status_code == status_pb2.Status.OK:
                        # update deployment state
                        if response and response.deployment:
                            self.db.deployment_store.insert_or_update(
                                sess, response.deployment
                            )
                        else:
                            raise BentoMLException(
                                "DeploymentOperator Internal Error: failed to add or "
                                "update deployment metadata to database"
                            )
                        logger.info(
                            "ApplyDeployment (%s, namespace %s) succeeded",
                            request.deployment.name,
                            request.deployment.namespace,
                        )
                    else:
                        if not previous_deployment:
                            # When failed to create the deployment,
                            # delete it from active deployments records
                            self.db.deployment_store.delete(
                                sess,
                                request.deployment.name,
                                request.deployment.namespace,
                            )
                        logger.debug(
                            "ApplyDeployment (%s, namespace %s) failed: %s",
                            request.deployment.name,
                            request.deployment.namespace,
                            response.status.error_message,
                        )

                    return response

                except BentoMLException as e:
                    logger.error("RPC ERROR ApplyDeployment: %s", e)
                    return ApplyDeploymentResponse(status=e.status_proto)
                except Exception as e:
                    logger.error("URPC ERROR ApplyDeployment: %s", e)
                    return ApplyDeploymentResponse(status=Status.INTERNAL(str(e)))

        def DeleteDeployment(self, request, context=None):
            deployment_id = f"{request.deployment_name}_{request.namespace}"
            with lock(self.db, [(deployment_id, LockType.WRITE)]) as (
                sess,
                _,
            ):
                try:
                    request.namespace = request.namespace or self.default_namespace
                    deployment_pb = self.db.deployment_store.get(
                        sess, request.deployment_name, request.namespace
                    )

                    if deployment_pb:
                        # find deployment operator based on deployment spec
                        operator = get_deployment_operator(self, deployment_pb)

                        # executing deployment deletion
                        response = operator.delete(deployment_pb)

                        # if delete successful, remove it from active deployment records
                        # table
                        if response.status.status_code == status_pb2.Status.OK:
                            track_deployment_delete(
                                deployment_pb.spec.operator, deployment_pb.created_at
                            )
                            self.db.deployment_store.delete(
                                sess, request.deployment_name, request.namespace
                            )
                            return response

                        # If force delete flag is True, we will remove the record
                        # from yatai database.
                        if request.force_delete:
                            # Track deployment delete before it
                            # vanishes from deployment store
                            track_deployment_delete(
                                deployment_pb.spec.operator,
                                deployment_pb.created_at,
                                True,
                            )
                            self.db.deployment_store.delete(
                                sess, request.deployment_name, request.namespace
                            )
                            return DeleteDeploymentResponse(status=Status.OK())

                        if response.status.status_code == status_pb2.Status.NOT_FOUND:
                            modified_message = (
                                'Cloud resources not found, error: {} - it '
                                'may have been deleted manually. Try delete '
                                'deployment with "--force" option to ignore this '
                                'error and force deleting the deployment record'.format(
                                    response.status.error_message
                                )
                            )
                            response.status.error_message = modified_message

                        return response
                    else:
                        return DeleteDeploymentResponse(
                            status=Status.NOT_FOUND(
                                'Deployment "{}" in namespace "{}" not found'.format(
                                    request.deployment_name, request.namespace
                                )
                            )
                        )

                except BentoMLException as e:
                    logger.error("RPC ERROR DeleteDeployment: %s", e)
                    return DeleteDeploymentResponse(status=e.status_proto)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error("RPC ERROR DeleteDeployment: %s", e)
                    return DeleteDeploymentResponse(status=Status.INTERNAL(str(e)))

        def GetDeployment(self, request, context=None):
            deployment_id = f"{request.deployment_name}_{request.namespace}"
            with lock(self.db, [(deployment_id, LockType.READ)]) as (
                sess,
                _,
            ):
                try:
                    request.namespace = request.namespace or self.default_namespace
                    deployment_pb = self.db.deployment_store.get(
                        sess, request.deployment_name, request.namespace
                    )
                    if deployment_pb:
                        return GetDeploymentResponse(
                            status=Status.OK(), deployment=deployment_pb
                        )
                    else:
                        return GetDeploymentResponse(
                            status=Status.NOT_FOUND(
                                'Deployment "{}" in namespace "{}" not found'.format(
                                    request.deployment_name, request.namespace
                                )
                            )
                        )
                except BentoMLException as e:
                    logger.error("RPC ERROR GetDeployment: %s", e)
                    return GetDeploymentResponse(status=e.status_proto)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error("RPC ERROR GetDeployment: %s", e)
                    return GetDeploymentResponse(status=Status.INTERNAL())

        def DescribeDeployment(self, request, context=None):
            deployment_id = f"{request.deployment_name}_{request.namespace}"
            with lock(self.db, [(deployment_id, LockType.READ)]) as (
                sess,
                _,
            ):
                try:
                    request.namespace = request.namespace or self.default_namespace
                    deployment_pb = self.db.deployment_store.get(
                        sess, request.deployment_name, request.namespace
                    )

                    if deployment_pb:
                        operator = get_deployment_operator(self, deployment_pb)

                        response = operator.describe(deployment_pb)

                        if response.status.status_code == status_pb2.Status.OK:
                            with self.db.deployment_store.update_deployment(
                                sess, request.deployment_name, request.namespace
                            ) as deployment:
                                deployment.state = ProtoMessageToDict(response.state)

                        return response
                    else:
                        return DescribeDeploymentResponse(
                            status=Status.NOT_FOUND(
                                'Deployment "{}" in namespace "{}" not found'.format(
                                    request.deployment_name, request.namespace
                                )
                            )
                        )
                except BentoMLException as e:
                    logger.error("RPC ERROR DescribeDeployment: %s", e)
                    return DeleteDeploymentResponse(status=e.status_proto)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error("RPC ERROR DescribeDeployment: %s", e)
                    return DeleteDeploymentResponse(status=Status.INTERNAL())

        def ListDeployments(self, request, context=None):
            with self.db.create_session() as sess:
                try:
                    namespace = request.namespace or self.default_namespace
                    deployment_pb_list = self.db.deployment_store.list(
                        sess,
                        namespace=namespace,
                        label_selectors=request.label_selectors,
                        offset=request.offset,
                        limit=request.limit,
                        operator=request.operator,
                        order_by=request.order_by,
                        ascending_order=request.ascending_order,
                    )

                    return ListDeploymentsResponse(
                        status=Status.OK(), deployments=deployment_pb_list
                    )
                except BentoMLException as e:
                    logger.error("RPC ERROR ListDeployments: %s", e)
                    return ListDeploymentsResponse(status=e.status_proto)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error("RPC ERROR ListDeployments: %s", e)
                    return ListDeploymentsResponse(status=Status.INTERNAL())

        def AddBento(self, request, context=None):
            bento_id = f"{request.bento_name}_{request.bento_version}"
            with lock(self.db, [(bento_id, LockType.WRITE)]) as (sess, _):
                try:
                    # TODO: validate request
                    bento_pb = self.db.metadata_store.get(
                        sess, request.bento_name, request.bento_version
                    )
                    if bento_pb:
                        error_message = (
                            "BentoService bundle: "
                            "{}:{} already exists".format(
                                request.bento_name, request.bento_version
                            )
                        )
                        logger.error(error_message)
                        return AddBentoResponse(status=Status.ABORTED(error_message))
                    new_bento_uri = self.repo.add(
                        request.bento_name, request.bento_version
                    )
                    self.db.metadata_store.add(
                        sess,
                        bento_name=request.bento_name,
                        bento_version=request.bento_version,
                        uri=new_bento_uri.uri,
                        uri_type=new_bento_uri.type,
                    )
                    return AddBentoResponse(status=Status.OK(), uri=new_bento_uri)
                except BentoMLException as e:
                    logger.error("RPC ERROR AddBento: %s", e)
                    return DeleteDeploymentResponse(status=e.status_proto)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error("URPC ERROR AddBento: %s", e)
                    return DeleteDeploymentResponse(status=Status.INTERNAL())

        def UpdateBento(self, request, context=None):
            bento_id = f"{request.bento_name}_{request.bento_version}"
            with lock(self.db, [(bento_id, LockType.WRITE)]) as (sess, _):
                try:
                    # TODO: validate request
                    if request.upload_status:
                        self.db.metadata_store.update_upload_status(
                            sess,
                            request.bento_name,
                            request.bento_version,
                            request.upload_status,
                        )
                    if request.service_metadata:
                        self.db.metadata_store.update(
                            sess,
                            request.bento_name,
                            request.bento_version,
                            request.service_metadata,
                        )
                    return UpdateBentoResponse(status=Status.OK())
                except BentoMLException as e:
                    logger.error("RPC ERROR UpdateBento: %s", e)
                    return UpdateBentoResponse(status=e.status_proto)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error("RPC ERROR UpdateBento: %s", e)
                    return UpdateBentoResponse(status=Status.INTERNAL())

        def DangerouslyDeleteBento(self, request, context=None):
            bento_id = f"{request.bento_name}_{request.bento_version}"
            with lock(self.db, [(bento_id, LockType.WRITE)]) as (sess, _):
                try:
                    # TODO: validate request
                    bento_pb = self.db.metadata_store.get(
                        sess, request.bento_name, request.bento_version
                    )
                    if not bento_pb:
                        msg = (
                            f"BentoService "
                            f"{request.bento_name}:{request.bento_version} "
                            f"has already been deleted"
                        )
                        return DangerouslyDeleteBentoResponse(
                            status=Status.ABORTED(msg)
                        )

                    logger.debug(
                        'Deleting BentoService %s:%s',
                        request.bento_name,
                        request.bento_version,
                    )
                    self.db.metadata_store.dangerously_delete(
                        sess, request.bento_name, request.bento_version
                    )
                    self.repo.dangerously_delete(
                        request.bento_name, request.bento_version
                    )
                    return DangerouslyDeleteBentoResponse(status=Status.OK())
                except BentoMLException as e:
                    logger.error("RPC ERROR DangerouslyDeleteBento: %s", e)
                    return DangerouslyDeleteBentoResponse(status=e.status_proto)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error("RPC ERROR DangerouslyDeleteBento: %s", e)
                    return DangerouslyDeleteBentoResponse(status=Status.INTERNAL())

        def GetBento(self, request, context=None):
            bento_id = f"{request.bento_name}_{request.bento_version}"
            with lock(self.db, [(bento_id, LockType.READ)]) as (sess, _):
                try:
                    # TODO: validate request
                    bento_pb = self.db.metadata_store.get(
                        sess, request.bento_name, request.bento_version
                    )
                    if bento_pb:
                        if request.bento_version.lower() == 'latest':
                            logger.info(
                                'Getting latest version %s:%s',
                                request.bento_name,
                                bento_pb.version,
                            )
                        if bento_pb.uri.type == BentoUri.S3:
                            bento_pb.uri.s3_presigned_url = self.repo.get(
                                bento_pb.name, bento_pb.version
                            )
                        elif bento_pb.uri.type == BentoUri.GCS:
                            bento_pb.uri.gcs_presigned_url = self.repo.get(
                                bento_pb.name, bento_pb.version
                            )
                        return GetBentoResponse(status=Status.OK(), bento=bento_pb)
                    else:
                        return GetBentoResponse(
                            status=Status.NOT_FOUND(
                                "BentoService `{}:{}` is not found".format(
                                    request.bento_name, request.bento_version
                                )
                            )
                        )
                except BentoMLException as e:
                    logger.error("RPC ERROR GetBento: %s", e)
                    return GetBentoResponse(status=e.status_proto)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error("RPC ERROR GetBento: %s", e)
                    return GetBentoResponse(status=Status.INTERNAL())

        def ListBento(self, request, context=None):
            with self.db.create_session() as sess:
                try:
                    # TODO: validate request
                    bento_metadata_pb_list = self.db.metadata_store.list(
                        sess,
                        bento_name=request.bento_name,
                        offset=request.offset,
                        limit=request.limit,
                        order_by=request.order_by,
                        label_selectors=request.label_selectors,
                        ascending_order=request.ascending_order,
                    )

                    return ListBentoResponse(
                        status=Status.OK(), bentos=bento_metadata_pb_list
                    )
                except BentoMLException as e:
                    logger.error("RPC ERROR ListBento: %s", e)
                    return ListBentoResponse(status=e.status_proto)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error("RPC ERROR ListBento: %s", e)
                    return ListBentoResponse(status=Status.INTERNAL())

        def ContainerizeBento(self, request, context=None):
            bento_id = f"{request.bento_name}_{request.bento_version}"
            with lock(self.db, [(bento_id, LockType.READ)]) as (sess, _):
                try:
                    ensure_docker_available_or_raise()
                    tag = request.tag
                    if tag is None or len(tag) == 0:
                        name = to_valid_docker_image_name(request.bento_name)
                        version = to_valid_docker_image_version(request.bento_version)
                        tag = f"{name}:{version}"
                    if ":" not in tag:
                        version = to_valid_docker_image_version(request.bento_version)
                        tag = f"{tag}:{version}"
                    import docker

                    docker_client = docker.from_env()
                    bento_pb = self.db.metadata_store.get(
                        sess, request.bento_name, request.bento_version
                    )
                    if not bento_pb:
                        raise YataiRepositoryException(
                            f'BentoService '
                            f'{request.bento_name}:{request.bento_version} '
                            f'does not exist'
                        )

                    with TempDirectory() as temp_dir:
                        temp_bundle_path = f'{temp_dir}/{bento_pb.name}'
                        bento_service_bundle_path = bento_pb.uri.uri
                        if bento_pb.uri.type == BentoUri.S3:
                            bento_service_bundle_path = self.repo.get(
                                bento_pb.name, bento_pb.version
                            )
                        elif bento_pb.uri.type == BentoUri.GCS:
                            bento_service_bundle_path = self.repo.get(
                                bento_pb.name, bento_pb.version
                            )
                        safe_retrieve(bento_service_bundle_path, temp_bundle_path)
                        try:
                            resp = docker_client.api.build(
                                path=temp_bundle_path,
                                nocache=False,
                                tag=tag,
                                buildargs=dict(request.build_args),
                            )
                            docker_build_logs(resp)
                        except (
                            docker.errors.APIError,
                            docker.errors.BuildError,
                        ) as error:
                            logger.error(f'Encounter container building issue: {error}')
                            raise YataiRepositoryException(error)
                        if request.push is True:
                            try:
                                docker_client.images.push(
                                    repository=request.repository, tag=tag
                                )
                            except docker.errors.APIError as error:
                                raise YataiRepositoryException(error)
                        return ContainerizeBentoResponse(status=Status.OK(), tag=tag)
                except BentoMLException as e:
                    logger.error(f"RPC ERROR ContainerizeBento: {e}")
                    return ContainerizeBentoResponse(status=e.status_proto)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error(f"RPC ERROR ContainerizeBento: {e}")
                    return ContainerizeBentoResponse(status=Status.INTERNAL(e))

        def UploadBento(self, request_iterator, context=None):
            if not is_file_system_repo(self.repo):
                logger.error(
                    "UploadBento RPC only works with File System based repository, "
                    "for other types of repositories(s3, gcs, minio), "
                    "use pre-signed URL for upload"
                )
                return UploadBentoResponse(status=Status.INTERNAL(''))
            try:
                with self.db.create_session() as sess:
                    lock_obj = None
                    bento_pb = None
                    with TempDirectory() as temp_dir:
                        temp_tar_path = os.path.join(
                            temp_dir, f'{uuid.uuid4().hex[:12]}.tar'
                        )
                        file = open(temp_tar_path, 'wb+')
                        for request in request_iterator:
                            # Initial request is without bundle
                            if not request.bento_bundle:
                                bento_name = request.bento_name
                                bento_version = request.bento_version
                                bento_pb = self.db.metadata_store.get(
                                    sess, bento_name, bento_version
                                )
                                if not bento_pb:
                                    result_status = Status.NOT_FOUND(
                                        "BentoService `{}:{}` is not found".format(
                                            bento_name, bento_version
                                        )
                                    )
                                    return UploadBentoResponse(status=result_status)
                                if bento_pb.status:
                                    if bento_pb.status.status == UploadStatus.DONE:
                                        return UploadBentoResponse(
                                            status=Status.CANCELLED(
                                                f"Bento bundle `{bento_name}:"
                                                f"{bento_version}` is uploaded"
                                            )
                                        )
                                    if bento_pb.status.status == UploadStatus.UPLOADING:
                                        return UploadBentoResponse(
                                            status=Status.CANCELLED(
                                                f"Bento bundle `{bento_name}:"
                                                f"{bento_version}` is currently "
                                                f"uploading"
                                            )
                                        )
                                if lock_obj is None:
                                    lock_obj = LockStore.acquire(
                                        sess=sess,
                                        lock_type=LockType.WRITE,
                                        resource_id=f'{bento_name}_{bento_version}',
                                        ttl_min=DEFAULT_TTL_MIN,
                                    )
                            else:
                                if (
                                    bento_name == request.bento_name
                                    and bento_version == request.bento_version
                                ):
                                    file.write(request.bento_bundle)
                                else:
                                    lock_obj.release(sess)
                                    raise BadInput(
                                        f"Incoming stream request doesn't match "
                                        f"with initial request info "
                                        f"{bento_name}:{bento_version} - "
                                        f"{request.bento_name}:"
                                        f"{request.bento_version}"
                                    )
                        file.seek(0)
                        with tarfile.open(fileobj=file, mode='r') as tar:
                            tar.extractall(path=bento_pb.uri.uri)
                        upload_status = UploadStatus(status=UploadStatus.DONE)
                        upload_status.updated_at.GetCurrentTime()
                        self.db.metadata_store.update_upload_status(
                            sess, bento_name, bento_version, upload_status
                        )
                        lock_obj.release(sess)
                        return UploadBentoResponse(status=Status.OK())
            except BentoMLException as e:
                logger.error("RPC ERROR UploadBento: %s", e)
                return UploadBentoResponse(status=e.status_proto)
            except Exception as e:  # pylint: disable=broad-except
                logger.error("RPC ERROR UploadBento: %s", e)
                return UploadBentoResponse(status=Status.INTERNAL())
            finally:
                if file is not None:
                    file.close()

        def DownloadBento(self, request, context=None):
            if not is_file_system_repo(self.repo):
                logger.error(
                    "DownloadBento RPC only works with File System based repository, "
                    "for other types of repositories(s3, gcs, minio), "
                    "use pre-signed URL for download"
                )
                return DownloadBentoResponse(status=Status.INTERNAL(''))
            bento_id = f"{request.bento_name}_{request.bento_version}"
            with lock(self.db, [(bento_id, LockType.READ)]) as (sess, _):
                try:
                    bento_pb = self.db.metadata_store.get(
                        sess, request.bento_name, request.bento_version
                    )
                    responses_generator = DownloadBentoStreamResponses(
                        bento_name=request.bento_name,
                        bento_version=request.bento_version,
                        bento_bundle_path=bento_pb.uri.uri,
                    )
                    for response in responses_generator:
                        yield response
                except BentoMLException as e:
                    logger.error("RPC ERROR DownloadBento: %s", e)
                    return DownloadBentoResponse(status=e.status_proto)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error("RPC ERROR DownloadBento: %s", e)
                    return DownloadBentoResponse(status=Status.INTERNAL())
                finally:
                    responses_generator.close()

    # pylint: enable=unused-argument

    return YataiServiceImpl



# flake8: noqa
# pylint: skip-file
# Workaround for https://github.com/bentoml/BentoML/issues/984
# Issue 984 is largely fixed by enabling the conda `pip_interop_enabled` option,
# although this option is only available after conda version 4.6.0
# This import wrapper makes sure ruamel.yaml works properly when used with a docker base
# image that has installed conda version < 4.6.0
# See more details in https://github.com/bentoml/BentoML/pull/1012
try:
    from ruamel.yaml import YAML
except ModuleNotFoundError:
    from ruamel_yaml import YAML



import argparse
import json
import logging
import re
import sys
from typing import TYPE_CHECKING

import click
import psutil
from simple_di import Provide, inject

from bentoml import __version__
from bentoml.cli.click_utils import (
    BentoMLCommandGroup,
    CLI_COLOR_SUCCESS,
    _echo,
    conditional_argument,
)
from bentoml.cli.utils import Spinner
from bentoml.configuration.containers import BentoMLContainer
from bentoml.saved_bundle import (
    load_bento_service_api,
    load_bento_service_metadata,
    load_from_dir,
)
from bentoml.server import start_dev_server, start_prod_server
from bentoml.utils import ProtoMessageToDict, resolve_bundle_path
from bentoml.utils.docker_utils import validate_tag
from bentoml.utils.lazy_loader import LazyLoader
from bentoml.utils.open_api import get_open_api_spec_json
from bentoml.yatai.client import get_yatai_client

if TYPE_CHECKING:
    from bentoml.yatai.proto.repository_pb2 import BentoServiceMetadata
    from bentoml.yatai.client import YataiClient

try:
    import click_completion

    click_completion.init()
    shell_types = click_completion.DocumentedChoice(click_completion.core.shells)
except ImportError:
    # click_completion package is optional to use BentoML cli,
    click_completion = None
    shell_types = click.Choice(['bash', 'zsh', 'fish', 'powershell'])

logger = logging.getLogger(__name__)

yatai_proto = LazyLoader('yatai_proto', globals(), 'bentoml.yatai.proto')


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def escape_shell_params(param):
    k, v = param.split("=")
    v = re.sub(r"([^a-zA-Z0-9])", r"\\\1", v)
    return "{}={}".format(k, v)


@inject
def create_bento_service_cli(
    pip_installed_bundle_path: str = None,
    default_yatai_url: str = None,
    default_port: int = Provide[BentoMLContainer.config.bento_server.port],
    default_mb_max_batch_size: int = Provide[
        BentoMLContainer.config.bento_server.microbatch.max_batch_size
    ],
    default_mb_max_latency: int = Provide[
        BentoMLContainer.config.bento_server.microbatch.max_latency
    ],
    default_run_with_ngrok: bool = Provide[
        BentoMLContainer.config.bento_server.ngrok.enabled
    ],
    default_enable_swagger: bool = Provide[
        BentoMLContainer.config.bento_server.swagger.enabled
    ],
    default_workers: int = Provide[BentoMLContainer.config.bento_server.workers],
    default_microbatch_workers: int = Provide[
        BentoMLContainer.config.bento_server.microbatch.workers
    ],
    default_timeout: int = Provide[BentoMLContainer.config.bento_server.timeout],
):
    # pylint: disable=unused-variable

    batch_options = [
        click.option(
            '--enable-microbatch/--disable-microbatch',
            default=None,
            help="Deprecated option for running API server with micro-batch enabled.",
            envvar='BENTOML_ENABLE_MICROBATCH',
        ),
        click.option(
            '--mb-max-batch-size',
            type=click.INT,
            help="Specify micro batching maximal batch size.",
            envvar='BENTOML_MB_MAX_BATCH_SIZE',
            default=default_mb_max_batch_size,
        ),
        click.option(
            '--mb-max-latency',
            type=click.INT,
            help="Specify micro batching maximal latency in milliseconds.",
            envvar='BENTOML_MB_MAX_LATENCY',
            default=default_mb_max_latency,
        ),
    ]

    @click.group(cls=BentoMLCommandGroup)
    @click.version_option(version=__version__)
    def bentoml_cli():
        """
        BentoML CLI tool
        """

    # Example Usage: bentoml run {API_NAME} {BUNDLE_PATH} --input ...
    @bentoml_cli.command(
        help="Run a API defined in saved BentoService bundle from command line",
        short_help="Run API function",
        context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
    )
    @conditional_argument(pip_installed_bundle_path is None, "bento", type=click.STRING)
    @click.argument("api_name", type=click.STRING)
    @click.argument('run_args', nargs=-1, type=click.UNPROCESSED)
    def run(api_name, run_args, bento=None):
        parser = argparse.ArgumentParser()
        parser.add_argument('--yatai-url', type=str, default=default_yatai_url)
        parsed_args, _ = parser.parse_known_args(run_args)
        yatai_url = parsed_args.yatai_url
        saved_bundle_path = resolve_bundle_path(
            bento, pip_installed_bundle_path, yatai_url
        )

        api = load_bento_service_api(saved_bundle_path, api_name)
        exit_code = api.handle_cli(run_args)
        sys.exit(exit_code)

    # Example Usage: bentoml info {BUNDLE_PATH}
    @bentoml_cli.command(
        help="List all APIs defined in the BentoService loaded from saved bundle",
        short_help="List APIs",
    )
    @conditional_argument(pip_installed_bundle_path is None, "bento", type=click.STRING)
    @click.option(
        '--yatai-url',
        type=click.STRING,
        default=default_yatai_url,
        help='Remote YataiService URL. Optional. '
        'Example: "--yatai-url http://localhost:50050"',
    )
    def info(bento=None, yatai_url=None):
        """
        List all APIs defined in the BentoService loaded from saved bundle
        """
        saved_bundle_path = resolve_bundle_path(
            bento, pip_installed_bundle_path, yatai_url
        )

        bento_service_metadata_pb = load_bento_service_metadata(saved_bundle_path)
        output = json.dumps(ProtoMessageToDict(bento_service_metadata_pb), indent=2)
        _echo(output)

    # Example usage: bentoml open-api-spec {BUNDLE_PATH}
    @bentoml_cli.command(
        name="open-api-spec",
        help="Display API specification JSON in Open-API format",
        short_help="Display OpenAPI/Swagger JSON specs",
    )
    @conditional_argument(pip_installed_bundle_path is None, "bento", type=click.STRING)
    @click.option(
        '--yatai-url',
        type=click.STRING,
        default=default_yatai_url,
        help='Remote YataiService URL. Optional. '
        'Example: "--yatai-url http://localhost:50050"',
    )
    def open_api_spec(bento=None, yatai_url=None):
        saved_bundle_path = resolve_bundle_path(
            bento, pip_installed_bundle_path, yatai_url
        )

        bento_service = load_from_dir(saved_bundle_path)

        _echo(json.dumps(get_open_api_spec_json(bento_service), indent=2))

    # Example Usage: bentoml serve {BUNDLE_PATH} --port={PORT}
    @bentoml_cli.command(
        help="Start a dev API server serving specified BentoService",
        short_help="Start local dev API server",
    )
    @conditional_argument(pip_installed_bundle_path is None, "bento", type=click.STRING)
    @click.option(
        "--port",
        type=click.INT,
        default=default_port,
        help="The port to listen on for the REST api server, default is 5000",
        envvar='BENTOML_PORT',
    )
    @add_options(batch_options)
    @click.option(
        '--run-with-ngrok',
        is_flag=True,
        default=default_run_with_ngrok,
        help="Use ngrok to relay traffic on a public endpoint to this "
        "API server on localhost",
        envvar='BENTOML_ENABLE_NGROK',
    )
    @click.option(
        '--yatai-url',
        type=click.STRING,
        default=default_yatai_url,
        help='Remote YataiService URL. Optional. '
        'Example: "--yatai-url http://localhost:50050"',
    )
    @click.option(
        '--enable-swagger/--disable-swagger',
        is_flag=True,
        default=default_enable_swagger,
        help="Run API server with Swagger UI enabled",
        envvar='BENTOML_ENABLE_SWAGGER',
    )
    def serve(
        port,
        bento,
        enable_microbatch,
        mb_max_batch_size,
        mb_max_latency,
        run_with_ngrok,
        yatai_url,
        enable_swagger,
    ):

        if enable_microbatch is not None:
            logger.warning(
                "Option --enable-microbatch/--disable-microbatch has been "
                "deprecated in the current release. The micro-batching option "
                "has become the default. Consider using --mb-max-batching=1 "
                "to simulate the effect of --disable-microbatch"
            )

        saved_bundle_path = resolve_bundle_path(
            bento, pip_installed_bundle_path, yatai_url
        )

        start_dev_server(
            saved_bundle_path,
            port=port,
            mb_max_batch_size=mb_max_batch_size,
            mb_max_latency=mb_max_latency,
            run_with_ngrok=run_with_ngrok,
            enable_swagger=enable_swagger,
        )

    # Example Usage:
    # bentoml serve-gunicorn {BUNDLE_PATH} --port={PORT} --workers={WORKERS}
    @bentoml_cli.command(
        help="Start a production API server serving specified BentoService",
        short_help="Start production API server",
    )
    @conditional_argument(pip_installed_bundle_path is None, "bento", type=click.STRING)
    @click.option(
        "-p",
        "--port",
        type=click.INT,
        default=default_port,
        help="The port to listen on for the REST api server, default is 5000",
        envvar='BENTOML_PORT',
    )
    @click.option(
        "-w",
        "--workers",
        type=click.INT,
        default=default_workers,
        help="Number of workers will start for the gunicorn server",
        envvar='BENTOML_GUNICORN_WORKERS',
    )
    @click.option(
        "--timeout",
        type=click.INT,
        default=default_timeout,
        help="Number of seconds for an API server request to timeout",
    )
    @add_options(batch_options)
    @click.option(
        '--microbatch-workers',
        type=click.INT,
        default=default_microbatch_workers,
        help="Number of micro-batch request dispatcher workers",
        envvar='BENTOML_MICROBATCH_WORKERS',
    )
    @click.option(
        '--yatai-url',
        type=click.STRING,
        default=default_yatai_url,
        help='Remote YataiService URL. Optional. '
        'Example: "--yatai-url http://localhost:50050"',
    )
    @click.option(
        '--enable-swagger/--disable-swagger',
        is_flag=True,
        default=default_enable_swagger,
        help="Run API server with Swagger UI enabled",
        envvar='BENTOML_ENABLE_SWAGGER',
    )
    def serve_gunicorn(
        port,
        workers,
        timeout,
        bento,
        enable_microbatch,
        mb_max_batch_size,
        mb_max_latency,
        microbatch_workers,
        yatai_url,
        enable_swagger,
    ):
        if not psutil.POSIX:
            _echo(
                "The `bentoml serve-gunicorn` command is only supported on POSIX. "
                "On windows platform, use `bentoml serve` for local API testing and "
                "docker for running production API endpoint: "
                "https://docs.docker.com/docker-for-windows/ "
            )
            return

        if enable_microbatch is not None:
            logger.warning(
                "Option --enable-microbatch/--disable-microbatch has been "
                "deprecated in the current release. The micro-batching option "
                "has become the default. Consider using --mb-max-batching=1 "
                "to simulate the effect of --disable-microbatch"
            )
        saved_bundle_path = resolve_bundle_path(
            bento, pip_installed_bundle_path, yatai_url
        )

        start_prod_server(
            saved_bundle_path,
            port=port,
            workers=workers,
            timeout=timeout,
            enable_swagger=enable_swagger,
            mb_max_batch_size=mb_max_batch_size,
            mb_max_latency=mb_max_latency,
            microbatch_workers=microbatch_workers,
        )

    @bentoml_cli.command(
        help="Install shell command completion",
        short_help="Install shell command completion",
    )
    @click.option(
        '--append/--overwrite',
        help="Append the completion code to the file",
        default=None,
    )
    @click.argument('shell', required=False, type=shell_types)
    @click.argument('path', required=False)
    def install_completion(append, shell, path):
        if click_completion:
            # click_completion package is imported
            shell, path = click_completion.core.install(
                shell=shell, path=path, append=append
            )
            click.echo('%s completion installed in %s' % (shell, path))
        else:
            click.echo(
                "'click_completion' is required for BentoML auto-completion, "
                "install it with `pip install click_completion`"
            )

    # bentoml containerize {service_name}:{service_tag} -t {docker_tags}
    @bentoml_cli.command(
        help='Containerizes given Bento into a ready-to-use Docker image.',
        short_help="Containerizes given Bento into a ready-to-use Docker image",
    )
    @click.argument("bento", type=click.STRING)
    @click.option('--push', is_flag=True)
    @click.option(
        '-t',
        '--tag',
        help="Optional image tag. If not specified, Bento will generate one from "
        "the name of the Bento.",
        required=False,
        callback=validate_tag,
    )
    @click.option(
        '--build-arg', multiple=True, help="pass through docker image build arguments"
    )
    @click.option(
        '--yatai-url',
        type=click.STRING,
        default=default_yatai_url,
        help='Specify the YataiService for running the containerization, default to '
        'the Local YataiService with local docker daemon. Example: '
        '"--yatai-url http://localhost:50050"',
    )
    def containerize(bento, push, tag, build_arg, yatai_url):
        """Containerize specified BentoService.

        BENTO is the target BentoService to be containerized, referenced by its name
        and version in format of name:version. For example: "iris_classifier:v1.2.0"

        `bentoml containerize` command also supports the use of the `latest` tag
        which will automatically use the last built version of your Bento.

        You can provide a tag for the image built by Bento using the
        `--tag` flag. Additionally, you can provide a `--push` flag,
        which will push the built image to the Docker repository specified by the
        image tag.

        You can also prefixing the tag with a hostname for the repository you wish
        to push to.
        e.g. `bentoml containerize IrisClassifier:latest --push --tag
        repo-address.com:username/iris` would build a Docker image called
        `username/iris:latest` and push that to docker repository at repo-address.com.

        By default, the `containerize` command will use the current credentials
        provided by Docker daemon.
        """
        saved_bundle_path = resolve_bundle_path(
            bento, pip_installed_bundle_path, yatai_url
        )

        _echo(f"Found Bento: {saved_bundle_path}")

        # fmt: off
        bento_metadata: "BentoServiceMetadata" = load_bento_service_metadata(saved_bundle_path)  # noqa: E501
        # fmt: on

        bento_tag = f'{bento_metadata.name}:{bento_metadata.version}'
        yatai_client: "YataiClient" = get_yatai_client(yatai_url)
        docker_build_args = {}
        if build_arg:
            for arg in build_arg:
                key, value = arg.split("=", 1)
                docker_build_args[key] = value
        if yatai_url is not None:
            spinner_message = f'Sending containerize RPC to YataiService at {yatai_url}'
        else:
            spinner_message = (
                f'Containerizing {bento_tag} with local YataiService and docker '
                f'daemon from local environment'
            )
        with Spinner(spinner_message):
            tag: str = yatai_client.repository.containerize(
                bento=bento_tag, tag=tag, build_args=docker_build_args, push=push,
            )
            _echo(f'\nBuild container image: {tag}', CLI_COLOR_SUCCESS)

    # pylint: enable=unused-variable
    return bentoml_cli



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import shutil
import re
import logging

from bentoml.utils.tempdir import TempDirectory
from bentoml.saved_bundle import load_bento_service_metadata
from bentoml.yatai.deployment.docker_utils import (
    ensure_docker_available_or_raise,
    build_docker_image,
)
from bentoml.adapters.clipper_input import ADAPTER_TYPE_TO_INPUT_TYPE
from bentoml.exceptions import BentoMLException

logger = logging.getLogger(__name__)


CLIPPER_ENTRY = """\
import rpc # this is clipper's rpc.py module
import os
import sys

from bentoml.saved_bundle import load_bento_service_api

IMPORT_ERROR_RETURN_CODE = 3


class BentoServiceContainer(rpc.ModelContainerBase):

    def __init__(self, bentoml_bundle_path, api_name):
        bento_service_api = load_bento_service_api(bentoml_bundle_path, api_name)
        self.predict_func = bento_service_api._func

    def predict_ints(self, inputs):
        preds = self.predict_func(inputs)
        return [str(p) for p in preds]

    def predict_floats(self, inputs):
        preds = self.predict_func(inputs)
        return [str(p) for p in preds]

    def predict_doubles(self, inputs):
        preds = self.predict_func(inputs)
        return [str(p) for p in preds]

    def predict_bytes(self, inputs):
        preds = self.predict_func(inputs)
        return [str(p) for p in preds]

    def predict_strings(self, inputs):
        preds = self.predict_func(inputs)
        return [str(p) for p in preds]


if __name__ == "__main__":
    print("Starting BentoService Clipper Container")
    rpc_service = rpc.RPCService()

    try:
        model = BentoServiceContainer('/container', '{api_name}')
        sys.stdout.flush()
        sys.stderr.flush()
    except ImportError:
        sys.exit(IMPORT_ERROR_RETURN_CODE)

    rpc_service.start(model)
"""


CLIPPER_DOCKERFILE = """\
FROM {base_image}

# copy over model files
# use /container to have access to python files in clipper base image such as rpc.py
COPY . /container
WORKDIR /container

# Install pip dependencies
ARG PIP_INDEX_URL=https://pypi.python.org/simple/
ARG PIP_TRUSTED_HOST=pypi.python.org
ENV PIP_INDEX_URL $PIP_INDEX_URL
ENV PIP_TRUSTED_HOST $PIP_TRUSTED_HOST

# Install conda, pip dependencies and run user defined setup script
RUN if [ -f /container/bentoml-init.sh ]; then bash -c /container/bentoml-init.sh; fi

ENV CLIPPER_MODEL_NAME={model_name}
ENV CLIPPER_MODEL_VERSION={model_version}

ENTRYPOINT []
CMD ["python", "/container/clipper_entry.py"]
"""  # noqa: E501


def get_clipper_compatible_string(item):
    """Generate clipper compatible string. It must be a valid DNS-1123.
    It must consist of lower case alphanumeric characters, '-' or '.',
    and must start and end with an alphanumeric character

    :param item: String
    :return: string
    """

    pattern = re.compile("[^a-zA-Z0-9-]")
    result = re.sub(pattern, "-", item)
    return result.lower()


def deploy_bentoml(
    clipper_conn, bundle_path, api_name, model_name=None, labels=None, build_envs=None
):
    """Deploy bentoml bundle to clipper cluster

    Args:
        clipper_conn(clipper_admin.ClipperConnection): Clipper connection instance
        bundle_path(str): Path to the saved BentomlService bundle.
        api_name(str): name of the api that will be used as prediction function for
            clipper cluster
        model_name(str): Model's name for clipper cluster
        labels(:obj:`list(str)`, optional): labels for clipper model

    Returns:
        tuple: Model name and model version that deployed to clipper

    """

    build_envs = {} if build_envs is None else build_envs
    # docker is required to build clipper model image
    ensure_docker_available_or_raise()

    if not clipper_conn.connected:
        raise BentoMLException(
            "No connection to Clipper cluster. CallClipperConnection.connect to "
            "connect to an existing cluster or ClipperConnection.start_clipper to "
            "create a new one"
        )

    bento_service_metadata = load_bento_service_metadata(bundle_path)

    try:
        api_metadata = next(
            (api for api in bento_service_metadata.apis if api.name == api_name)
        )
    except StopIteration:
        raise BentoMLException(
            "Can't find API '{}' in BentoService bundle {}".format(
                api_name, bento_service_metadata.name
            )
        )

    if api_metadata.input_type not in ADAPTER_TYPE_TO_INPUT_TYPE:
        raise BentoMLException(
            "Only BentoService APIs using ClipperInput can be deployed to Clipper"
        )

    input_type = ADAPTER_TYPE_TO_INPUT_TYPE[api_metadata.input_type]
    model_name = model_name or get_clipper_compatible_string(
        bento_service_metadata.name + "-" + api_metadata.name
    )
    model_version = get_clipper_compatible_string(bento_service_metadata.version)

    with TempDirectory() as tempdir:
        entry_py_content = CLIPPER_ENTRY.format(api_name=api_name)
        build_path = os.path.join(tempdir, "bento")
        shutil.copytree(bundle_path, build_path)

        with open(os.path.join(build_path, "clipper_entry.py"), "w") as f:
            f.write(entry_py_content)

        if bento_service_metadata.env.python_version.startswith("3.6"):
            base_image = "clipper/python36-closure-container:0.4.1"
        elif bento_service_metadata.env.python_version.startswith("2.7"):
            base_image = "clipper/python-closure-container:0.4.1"
        else:
            raise BentoMLException(
                "Python version {} is not supported in Clipper".format(
                    bento_service_metadata.env.python_version
                )
            )

        docker_content = CLIPPER_DOCKERFILE.format(
            model_name=model_name,
            model_version=model_version,
            base_image=base_image,
            pip_index_url=build_envs.get("PIP_INDEX_URL", ""),
            pip_trusted_url=build_envs.get("PIP_TRUSTED_HOST", ""),
        )
        with open(os.path.join(build_path, "Dockerfile-clipper"), "w") as f:
            f.write(docker_content)

        clipper_model_docker_image_tag = "clipper-model-{}:{}".format(
            bento_service_metadata.name.lower(), bento_service_metadata.version
        )
        build_docker_image(
            build_path, 'Dockerfile-clipper', clipper_model_docker_image_tag
        )

        logger.info(
            "Successfully built docker image %s for Clipper deployment",
            clipper_model_docker_image_tag,
        )

    clipper_conn.deploy_model(
        name=model_name,
        version=model_version,
        input_type=input_type,
        image=clipper_model_docker_image_tag,
        labels=labels,
    )

    return model_name, model_version



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import logging
import os
from pathlib import Path
from sys import version_info

from google.protobuf.struct_pb2 import Struct
from simple_di import Provide, inject

from bentoml import __version__ as BENTOML_VERSION
from bentoml.configuration.containers import BentoMLContainer
from bentoml.exceptions import BentoMLConfigException
from bentoml.utils import dump_to_yaml_str
from bentoml.utils.ruamel_yaml import YAML

BENTOML_CONFIG_YAML_TEMPLATE = """\
version: {bentoml_version}
kind: {kind}
metadata:
    created_at: {created_at}
"""

logger = logging.getLogger(__name__)
DEFAULT_MAX_LATENCY = 20000
DEFAULT_MAX_BATCH_SIZE = 4000
PYTHON_VERSION = f"{version_info.major}.{version_info.minor}.{version_info.micro}"


def _get_apis_list(bento_service):
    result = []
    for api in bento_service.inference_apis:
        api_obj = {
            "name": api.name,
            "docs": api.doc,
            "input_type": api.input_adapter.__class__.__name__,
            "output_type": api.output_adapter.__class__.__name__,
            "mb_max_batch_size": api.mb_max_batch_size,
            "mb_max_latency": api.mb_max_latency,
            "batch": api.batch,
            "route": api.route,
        }
        if api.input_adapter.config:
            api_obj["input_config"] = api.input_adapter.config
        if api.output_adapter.config:
            api_obj["output_config"] = api.output_adapter.config
        result.append(api_obj)
    return result


def _get_artifacts_list(bento_service):
    result = []
    for artifact_name, artifact in bento_service.artifacts.items():
        result.append(
            {
                'name': artifact_name,
                'artifact_type': artifact.__class__.__name__,
                'metadata': artifact.metadata,
            }
        )
    return result


class SavedBundleConfig(object):
    @inject
    def __init__(
        self,
        bento_service=None,
        kind="BentoService",
        bentoml_deployment_version: str = Provide[
            BentoMLContainer.bento_bundle_deployment_version
        ],
    ):
        self.kind = kind
        self._yaml = YAML()
        self._yaml.default_flow_style = False
        self.config = self._yaml.load(
            BENTOML_CONFIG_YAML_TEMPLATE.format(
                kind=self.kind,
                bentoml_version=bentoml_deployment_version,
                created_at=str(datetime.utcnow()),
            )
        )

        if bento_service is not None:
            self.config["metadata"].update(
                {
                    "service_name": bento_service.name,
                    "service_version": bento_service.version,
                }
            )
            self.config["env"] = bento_service.env.to_dict()
            self.config['apis'] = _get_apis_list(bento_service)
            self.config['artifacts'] = _get_artifacts_list(bento_service)

    def write_to_path(self, path, filename="bentoml.yml"):
        return self._yaml.dump(self.config, Path(os.path.join(path, filename)))

    @classmethod
    def load(cls, filepath):
        conf = cls()
        with open(filepath, "rb") as config_file:
            yml_content = config_file.read()
        conf.config = conf._yaml.load(yml_content)
        ver = str(conf["version"])
        py_ver = conf.config["env"]["python_version"]

        if ver != BENTOML_VERSION:
            msg = (
                "Saved BentoService bundle version mismatch: loading BentoService "
                "bundle create with BentoML version {}, but loading from BentoML "
                "version {}".format(conf["version"], BENTOML_VERSION)
            )

            # If major version is different, then there could be incompatible API
            # changes. Raise error in this case.
            if ver.split(".")[0] != BENTOML_VERSION.split(".")[0]:
                if not BENTOML_VERSION.startswith('0+untagged'):
                    raise BentoMLConfigException(msg)
                else:
                    logger.warning(msg)
            else:  # Otherwise just show a warning.
                logger.warning(msg)

        if py_ver != PYTHON_VERSION:
            logger.warning(
                f"Saved BentoService Python version mismatch: loading "
                f"BentoService bundle created with Python version {py_ver}, "
                f"but current environment version is {PYTHON_VERSION}."
            )

        return conf

    def get_bento_service_metadata_pb(self):
        from bentoml.yatai.proto.repository_pb2 import BentoServiceMetadata

        bento_service_metadata = BentoServiceMetadata()
        bento_service_metadata.name = self.config["metadata"]["service_name"]
        bento_service_metadata.version = self.config["metadata"]["service_version"]
        bento_service_metadata.created_at.FromDatetime(
            self.config["metadata"]["created_at"]
        )

        if "env" in self.config:
            env = self.config["env"]
            if "setup_sh" in env:
                bento_service_metadata.env.setup_sh = env["setup_sh"]

            if "conda_env" in env:
                bento_service_metadata.env.conda_env = dump_to_yaml_str(
                    env["conda_env"]
                )

            if "pip_packages" in env:
                for pip_package in env["pip_packages"]:
                    bento_service_metadata.env.pip_packages.append(pip_package)
            if "python_version" in env:
                bento_service_metadata.env.python_version = env["python_version"]
            if "docker_base_image" in env:
                bento_service_metadata.env.docker_base_image = env["docker_base_image"]
            if "requirements_txt" in env:
                bento_service_metadata.env.requirements_txt = env["requirements_txt"]

        if "apis" in self.config:
            for api_config in self.config["apis"]:
                if 'handler_type' in api_config:
                    # Convert handler type to input type for saved bundle created
                    # before version 0.8.0
                    input_type = api_config.get('handler_type')
                elif 'input_type' in api_config:
                    input_type = api_config.get('input_type')
                else:
                    input_type = "unknown"

                if 'output_type' in api_config:
                    output_type = api_config.get('output_type')
                else:
                    output_type = "DefaultOutput"

                api_metadata = BentoServiceMetadata.BentoServiceApi(
                    name=api_config["name"],
                    docs=api_config["docs"],
                    input_type=input_type,
                    output_type=output_type,
                )
                if "handler_config" in api_config:
                    # Supports viewing API input config info for saved bundle created
                    # before version 0.8.0
                    for k, v in api_config["handler_config"].items():
                        if k in {'mb_max_latency', 'mb_max_batch_size'}:
                            setattr(api_metadata, k, v)
                        else:
                            api_metadata.input_config[k] = v
                else:
                    if 'mb_max_latency' in api_config:
                        api_metadata.mb_max_latency = api_config["mb_max_latency"]
                    else:
                        api_metadata.mb_max_latency = DEFAULT_MAX_LATENCY

                    if 'mb_max_batch_size' in api_config:
                        api_metadata.mb_max_batch_size = api_config["mb_max_batch_size"]
                    else:
                        api_metadata.mb_max_batch_size = DEFAULT_MAX_BATCH_SIZE

                    if 'route' in api_config:
                        api_metadata.route = api_config["route"]
                    else:
                        # Use API name as the URL route when route config is missing,
                        # this is for backward compatibility for
                        # BentoML version <= 0.10.1
                        api_metadata.route = api_config["name"]

                if "input_config" in api_config:
                    for k, v in api_config["input_config"].items():
                        api_metadata.input_config[k] = v

                if "output_config" in api_config:
                    for k, v in api_config["output_config"].items():
                        api_metadata.output_config[k] = v
                api_metadata.batch = api_config.get("batch", False)
                bento_service_metadata.apis.extend([api_metadata])

        if "artifacts" in self.config:
            for artifact_config in self.config["artifacts"]:
                artifact_metadata = BentoServiceMetadata.BentoArtifact()
                if "name" in artifact_config:
                    artifact_metadata.name = artifact_config["name"]
                if "artifact_type" in artifact_config:
                    artifact_metadata.artifact_type = artifact_config["artifact_type"]
                if "metadata" in artifact_config:
                    if isinstance(artifact_config["metadata"], dict):
                        s = Struct()
                        s.update(artifact_config["metadata"])
                        artifact_metadata.metadata.CopyFrom(s)
                    else:
                        logger.warning(
                            "Tried to get non-dictionary metadata for artifact "
                            f"{artifact_metadata.name}. Ignoring metadata..."
                        )
                bento_service_metadata.artifacts.extend([artifact_metadata])

        return bento_service_metadata

    def __getitem__(self, item):
        return self.config[item]

    def __setitem__(self, key, value):
        self.config[key] = value

    def __contains__(self, item):
        return item in self.config



import time


class FixedBucket:
    '''
    Fixed size FIFO container.
    '''

    def __init__(self, size):
        self._data = [None] * size
        self._cur = 0
        self._size = size
        self._flag_full = False

    def put(self, v):
        self._data[self._cur] = v
        self._cur += 1
        if self._cur == self._size:
            self._cur = 0
            self._flag_full = True

    @property
    def data(self):
        if not self._flag_full:
            return self._data[: self._cur]
        return self._data

    def __len__(self):
        if not self._flag_full:
            return self._cur
        return self._size

    def __getitem__(self, sl):
        if not self._flag_full:
            return self._data[: self._cur][sl]
        return (self._data[self._cur :] + self._data[: self._cur])[sl]


class TokenBucket:
    '''
    Dynamic token bucket
    '''

    def __init__(self, init_amount=0):
        self._amount = init_amount
        self._last_consume_time = time.time()

    def consume(self, take_amount, avg_rate, burst_size):
        now = time.time()
        inc = (now - self._last_consume_time) * avg_rate
        current_amount = min(inc + self._amount, burst_size)
        if take_amount > current_amount:
            return False
        self._amount, self._last_consume_time = current_amount - take_amount, now
        return True



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

import click
from simple_di import Provide, inject

from bentoml.configuration.containers import (
    BentoMLContainer,
    YATAI_REPOSITORY_FILE_SYSTEM,
    YATAI_REPOSITORY_GCS,
    YATAI_REPOSITORY_S3,
    YATAI_REPOSITORY_TYPES,
)
from bentoml.yatai.yatai_service import start_yatai_service_grpc_server


logger = logging.getLogger(__name__)


@inject
def add_yatai_service_sub_command(
    cli,
    default_db_url: str = Provide[BentoMLContainer.yatai_database_url],
    default_repository_type: str = Provide[
        BentoMLContainer.config.yatai.repository.type
    ],
    default_file_system_directory: str = Provide[
        BentoMLContainer.yatai_file_system_directory
    ],
    default_s3_url: str = Provide[BentoMLContainer.config.yatai.repository.s3.url],
    default_s3_endpoint_url: str = Provide[
        BentoMLContainer.config.yatai.repository.s3.endpoint_url
    ],
    default_gcs_url: str = Provide[BentoMLContainer.config.yatai.repository.gcs.url],
):
    # pylint: disable=unused-variable

    @cli.command(help='Start BentoML YataiService for model management and deployment')
    @click.option(
        '--db-url',
        type=click.STRING,
        help='Database URL following RFC-1738, and usually can include username, '
        'password, hostname, database name as well as optional keyword arguments '
        'for additional configuration',
        envvar='BENTOML_DB_URL',
        default=default_db_url,
    )
    @click.option(
        '--repo-base-url',
        type=click.STRING,
        help='Deprecated option for base URL for storing BentoML saved bundle files, '
        'this can be a file system path(POSIX/Windows), or a S3 URL, '
        'usually starting with "s3://"',
        envvar='BENTOML_REPO_BASE_URL',
    )
    @click.option(
        '--grpc-port',
        type=click.INT,
        default=50051,
        help='Port to run YataiService gRPC server',
        envvar='BENTOML_GRPC_PORT',
    )
    @click.option(
        '--ui-port',
        type=click.INT,
        default=3000,
        help='Port to run YataiService Web UI server',
        envvar='BENTOML_WEB_UI_PORT',
    )
    @click.option(
        '--ui/--no-ui',
        default=True,
        help='Run YataiService with or without Web UI, when running with --no-ui, it '
        'will only run the gRPC server',
        envvar='BENTOML_ENABLE_WEB_UI',
    )
    @click.option(
        '--web-prefix-path',
        type=click.STRING,
        default='.',
        help='Add a location prefix to the URL when running YataiService'
        'behind a reverse proxy server',
        envvar='BENTOML_YATAI_WEB_PREFIX_PATH',
    )
    @click.option(
        '--repository-type',
        type=click.Choice(YATAI_REPOSITORY_TYPES, case_sensitive=False),
        default=default_repository_type,
        help='Type of the repository implementation',
        envvar='BENTOML_YATAI_REPOSITORY_TYPE',
    )
    @click.option(
        '--file-system-directory',
        type=click.STRING,
        default=default_file_system_directory,
        help='Specifies the directory path for the file system repository type',
        envvar='BENTOML_YATAI_FILE_SYSTEM_DIRECTORY',
    )
    @click.option(
        '--s3-url',
        type=click.STRING,
        default=default_s3_url,
        help='Specifies the S3 URL for the S3 repository type',
        envvar='BENTOML_YATAI_S3_URL',
    )
    @click.option(
        '--s3-endpoint-url',
        type=click.STRING,
        default=default_s3_endpoint_url,
        help='S3 Endpoint URL is used for deploying with storage services that are '
        'compatible with Amazon S3, such as MinIO',
        envvar='BENTOML_S3_ENDPOINT_URL',
    )
    @click.option(
        '--gcs-url',
        type=click.STRING,
        default=default_gcs_url,
        help='Specifies the GCS URL for the GCS repository type',
        envvar='BENTOML_YATAI_GCS_URL',
    )
    def yatai_service_start(
        db_url,
        repo_base_url,
        grpc_port,
        ui_port,
        ui,
        web_prefix_path,
        repository_type,
        file_system_directory,
        s3_url,
        s3_endpoint_url,
        gcs_url,
    ):
        from bentoml.utils.s3 import is_s3_url
        from bentoml.utils.gcs import is_gcs_url

        if repo_base_url:
            logger.warning(
                "Option --repo-base-url has been deprecated but is still supported "
                "in the current release. Consider using --repository-type and its "
                "corresponding options in the upcoming releases. "
            )
            if is_s3_url(repo_base_url):
                repository_type = YATAI_REPOSITORY_S3
                s3_url = repo_base_url
            elif is_gcs_url(repo_base_url):
                repository_type = YATAI_REPOSITORY_GCS
                gcs_url = repo_base_url
            else:
                repository_type = YATAI_REPOSITORY_FILE_SYSTEM
                file_system_directory = repo_base_url

        if repository_type == YATAI_REPOSITORY_S3 and s3_url is None:
            logger.error("'--s3-url' must be specified for S3 repository type")
            return
        elif repository_type == YATAI_REPOSITORY_GCS and gcs_url is None:
            logger.error("'--gcs-url' must be specified for GCS repository type")
            return
        elif (
            repository_type == YATAI_REPOSITORY_FILE_SYSTEM
            and file_system_directory is None
        ):
            logger.error(
                "'--file-system-directory' must be specified for file system "
                "repository type"
            )
            return
        else:
            start_yatai_service_grpc_server(
                db_url=db_url,
                grpc_port=grpc_port,
                ui_port=ui_port,
                with_ui=ui,
                base_url=web_prefix_path,
                repository_type=repository_type,
                file_system_directory=file_system_directory,
                s3_url=s3_url,
                s3_endpoint_url=s3_endpoint_url,
                gcs_url=gcs_url,
            )



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import logging
import re
import tarfile
from pathlib import Path

import boto3
from botocore.exceptions import ClientError

from bentoml.exceptions import BentoMLException
from bentoml.yatai.deployment.aws_utils import call_sam_command

logger = logging.getLogger(__name__)

# Maximum size for Lambda function bundle 250MB, leaving 1mb offset
LAMBDA_FUNCTION_LIMIT = 249000000
# Total disk size that user has access to.  Lambda function bundle + /tmp 512MB, leaving
# 1mb offset
LAMBDA_TEMPORARY_DIRECTORY_MAX_LIMIT = 511000000
LAMBDA_FUNCTION_MAX_LIMIT = LAMBDA_FUNCTION_LIMIT + LAMBDA_TEMPORARY_DIRECTORY_MAX_LIMIT


def init_sam_project(
    sam_project_path,
    bento_service_bundle_path,
    deployment_name,
    bento_name,
    api_names,
    aws_region,
):
    function_path = os.path.join(sam_project_path, deployment_name)
    os.mkdir(function_path)
    # Copy requirements.txt
    logger.debug("Copying requirements.txt")
    requirement_txt_path = os.path.join(bento_service_bundle_path, "requirements.txt")
    shutil.copy(requirement_txt_path, function_path)

    bundled_dep_path = os.path.join(
        bento_service_bundle_path, "bundled_pip_dependencies"
    )
    if os.path.isdir(bundled_dep_path):
        # Copy bundled pip dependencies
        logger.debug("Copying bundled_dependencies")
        shutil.copytree(
            bundled_dep_path, os.path.join(function_path, "bundled_pip_dependencies")
        )
        bundled_files = os.listdir(
            os.path.join(function_path, "bundled_pip_dependencies")
        )
        for index, bundled_file_name in enumerate(bundled_files):
            bundled_files[index] = "./bundled_pip_dependencies/{}\n".format(
                bundled_file_name
            )

        logger.debug("Updating requirements.txt")
        with open(
            os.path.join(function_path, "requirements.txt"), "r+"
        ) as requirement_file:
            required_modules = bundled_files + requirement_file.readlines()
            # Write from beginning of the file, instead of appending to the end.
            requirement_file.seek(0)
            requirement_file.writelines(required_modules)

    # Copy bento_service_model
    logger.debug("Copying model directory")
    model_path = os.path.join(bento_service_bundle_path, bento_name)
    shutil.copytree(model_path, os.path.join(function_path, bento_name))

    logger.debug("Creating python files for lambda function")
    # Create empty __init__.py
    open(os.path.join(function_path, "__init__.py"), "w").close()

    app_py_path = os.path.join(os.path.dirname(__file__), "lambda_app.py")
    shutil.copy(app_py_path, os.path.join(function_path, "app.py"))
    unzip_requirement_py_path = os.path.join(
        os.path.dirname(__file__), "download_extra_resources.py"
    )
    shutil.copy(
        unzip_requirement_py_path,
        os.path.join(function_path, "download_extra_resources.py"),
    )

    logger.info("Building lambda project")
    return_code, stdout, stderr = call_sam_command(
        ["build", "--use-container", "--region", aws_region],
        project_dir=sam_project_path,
        region=aws_region,
    )
    if return_code != 0:
        error_message = stderr
        if not error_message:
            error_message = stdout
        raise BentoMLException(
            "Failed to build lambda function. {}".format(error_message)
        )
    logger.debug("Removing unnecessary files to free up space")
    for api_name in api_names:
        cleanup_build_files(sam_project_path, api_name)


def cleanup_build_files(project_dir, api_name):
    build_dir = os.path.join(project_dir, ".aws-sam/build/{}".format(api_name))
    logger.debug("Cleaning up unused files in SAM built directory %s", build_dir)
    for root, dirs, files in os.walk(build_dir):
        if "tests" in dirs:
            logger.debug("removing dir: %s", os.path.join(root, "tests"))
            shutil.rmtree(os.path.join(root, "tests"))
        if "__pycache__" in dirs:
            logger.debug("removing dir: %s", os.path.join(root, "__pycache__"))
            shutil.rmtree(os.path.join(root, "__pycache__"))

        for dir_name in filter(lambda x: re.match("^.*\\.dist-info$", x), dirs):
            logger.debug("removing dir: %s", dir_name)
            shutil.rmtree(os.path.join(root, dir_name))

        for file in filter(lambda x: re.match("^.*\\.egg-info$", x), files):
            logger.debug("removing file: %s", os.path.join(root, file))
            os.remove(os.path.join(root, file))

        for file in filter(lambda x: re.match("^.*\\.pyc$", x), files):
            logger.debug("removing file: %s", os.path.join(root, file))
            os.remove(os.path.join(root, file))

        if "wheel" in files:
            logger.debug("removing file: %s", os.path.join(root, "wheel"))
            os.remove(os.path.join(root, "wheel"))
        if "pip" in files:
            logger.debug("removing file: %s", os.path.join(root, "pip"))
            os.remove(os.path.join(root, "pip"))
        if "caff2" in files:
            logger.debug("removing file: %s", os.path.join(root, "caff2"))
            os.remove(os.path.join(root, "caff2"))


def lambda_package(project_dir, aws_region, s3_bucket_name, deployment_prefix):
    prefix_path = os.path.join(deployment_prefix, "lambda-functions")
    build_dir = os.path.join(project_dir, ".aws-sam", "build")

    return_code, stdout, stderr = call_sam_command(
        [
            "package",
            "--force-upload",
            "--s3-bucket",
            s3_bucket_name,
            "--s3-prefix",
            prefix_path,
            "--template-file",
            "template.yaml",
            "--output-template-file",
            "packaged.yaml",
            "--region",
            aws_region,
        ],
        project_dir=build_dir,
        region=aws_region,
    )
    if return_code != 0:
        error_message = stderr
        if not error_message:
            error_message = stdout
        raise BentoMLException(
            "Failed to package lambda function. {}".format(error_message)
        )
    else:
        return stdout


def lambda_deploy(project_dir, aws_region, stack_name):
    # if the stack name exists and the state is in rollback_complete or
    # other 'bad' state, we will delete the stack first, and then deploy
    # it
    logger.debug('Ensure stack "%s" is ready to deploy', stack_name)
    ensure_is_ready_to_deploy_to_cloud_formation(stack_name, aws_region)
    logger.debug('Stack "%s"is ready to deploy', stack_name)

    template_file = os.path.join(project_dir, ".aws-sam", "build", "packaged.yaml")
    return_code, stdout, stderr = call_sam_command(
        [
            "deploy",
            "--stack-name",
            stack_name,
            "--capabilities",
            "CAPABILITY_IAM",
            "--template-file",
            template_file,
            "--region",
            aws_region,
        ],
        project_dir=project_dir,
        region=aws_region,
    )
    if return_code != 0:
        error_message = stderr
        if not error_message:
            error_message = stdout
        raise BentoMLException(
            "Failed to deploy lambda function. {}".format(error_message)
        )
    else:
        return stdout


def total_file_or_directory_size(file_or_dir):
    if os.path.isdir(file_or_dir):
        return sum(
            f.stat().st_size for f in Path(file_or_dir).glob("**/*") if f.is_file()
        )
    else:
        return Path(file_or_dir).stat().st_size


def reduce_bundle_size_and_upload_extra_resources_to_s3(
    build_directory,
    region,
    s3_bucket,
    deployment_prefix,
    function_name,
    lambda_project_dir,
):
    additional_requirements_path = os.path.join(
        lambda_project_dir, "additional_requirements", function_name
    )
    upload_dir_path = os.path.join(additional_requirements_path, "requirements")
    os.makedirs(upload_dir_path, exist_ok=True)
    s3_client = boto3.client("s3", region)

    dir_name_to_size = dict(
        (item, total_file_or_directory_size(os.path.join(build_directory, item)))
        for item in os.listdir(build_directory)
    )

    required_bundle_list = ["app.py", "__init__.py", "download_extra_resources.py"]
    required_bundle_size = sum(dir_name_to_size[i] for i in required_bundle_list)
    for name, size in sorted(
        dir_name_to_size.items(), key=lambda i: i[1], reverse=True
    ):
        if name in required_bundle_list:
            logger.debug('Including "%s" it is part of required bundle.', name)
            continue
        else:
            new_bundle_size = required_bundle_size + size
            if new_bundle_size >= LAMBDA_FUNCTION_LIMIT:
                logger.debug(
                    'Lambda function size %d is over the limit. Moving item "%s" '
                    "out of the bundle directory",
                    new_bundle_size,
                    name,
                )
                package_path = os.path.join(build_directory, name)
                copy_dst_path = os.path.join(upload_dir_path, name)
                shutil.move(package_path, copy_dst_path)
            else:
                logger.debug(
                    'Including "%s" Now the current lambda function size is %d',
                    name,
                    new_bundle_size,
                )
                required_bundle_size = new_bundle_size

    logger.debug("Final Lambda function bundle size is %d", required_bundle_size)
    additional_requirements_size = total_file_or_directory_size(
        additional_requirements_path
    )
    logger.debug("Additional requirement size is %d", additional_requirements_size)
    logger.debug("zip up additional requirement packages")
    tar_file_path = os.path.join(additional_requirements_path, "requirements.tar")
    with tarfile.open(tar_file_path, "w:gz") as tar:
        tar.add(upload_dir_path, arcname="requirements")
    logger.debug("Uploading requirements.tar to %s/%s", s3_bucket, deployment_prefix)
    try:
        s3_client.upload_file(
            tar_file_path,
            s3_bucket,
            os.path.join(deployment_prefix, "requirements.tar"),
        )
    except ClientError as e:
        raise BentoMLException(str(e))


def ensure_is_ready_to_deploy_to_cloud_formation(stack_name, region):
    try:
        cf_client = boto3.client("cloudformation", region)
        logger.debug("Checking stack description")
        describe_formation_result = cf_client.describe_stacks(StackName=stack_name)
        result_stacks = describe_formation_result.get("Stacks")
        if len(result_stacks):
            logger.debug('Stack "%s" exists', stack_name)
            stack_result = result_stacks[0]
            if stack_result["StackStatus"] in [
                "ROLLBACK_COMPLETE",
                "ROLLBACK_FAILED",
                "ROLLBACK_IN_PROGRESS",
            ]:
                logger.debug(
                    'Stack "%s" is in a "bad" status(%s), deleting the stack '
                    "before deployment",
                    stack_name,
                    stack_result["StackStatus"],
                )
                cf_client.delete_stack(StackName=stack_name)
    except ClientError as e:
        # We are brutally parse and handle stack doesn't exist, since
        # "AmazonCloudFormationException" currently is not implemented in boto3. Once
        # the current error is implemented, we need to switch
        error_response = e.response.get("Error", {})
        error_code = error_response.get("Code")
        error_message = error_response.get("Message", "Unknown")
        if error_code == "ValidationError" and "does not exist" in error_message:
            pass
        else:
            raise BentoMLException(str(e))
